# -*- coding: utf-8 -*-

from rest_framework import generics, status
from rest_framework.response import Response

from thux.organization.models import Organization

from . import serializers
from .. import models


class CartDetailView(generics.GenericAPIView):
    serializer_class = serializers.CartSerializer

    def get(self, request, *args, **kwargs):
        if 'cart_token' not in request.session:
            request.session['cart_token'] = models.Cart.create_cart_token()
        cart = models.Cart(cart_token=request.session['cart_token'], user=request.user)
        serializer = self.get_serializer(cart)
        return Response(serializer.data)


class CartRowDetailView(generics.GenericAPIView):
    serializer_class = serializers.CartRowCreateUpdateSerializer

    def get(self, request, *args, **kwargs):
        cart_row = models.CartRow(cart_token=request.session['cart_token'])
        serializer = self.get_serializer(cart_row)
        return Response(serializer.data)


class AddProductToCartView(generics.CreateAPIView):
    serializer_class = serializers.CartRowCreateUpdateSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        product = serializer.validated_data['product']
        quantity = serializer.validated_data['quantity']

        cart = models.Cart(cart_token=request.session['cart_token'], user=request.user)
        row = cart.add_product(product=product, quantity=quantity)
        serializer = self.get_serializer(row)
        return Response(serializer.data, status=status.HTTP_201_CREATED)


class UpdateProductCartView(generics.CreateAPIView):
    serializer_class = serializers.CartRowCreateUpdateSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        product_pk = serializer.validated_data['product'].pk
        quantity = serializer.validated_data['quantity']
        cart = models.Cart(cart_token=request.session['cart_token'], user=request.user)
        row = cart.update_product(product=product_pk, quantity=quantity)
        serializer = self.get_serializer(row)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class RemoveProductFromCartView(generics.DestroyAPIView):
    serializer_class = serializers.CartRowCreateUpdateSerializer

    def delete(self, request, *args, **kwargs):
        cart = models.Cart(cart_token=request.session['cart_token'], user=request.user)
        cart.delete_product(kwargs['pk'])
        return Response(status=status.HTTP_204_NO_CONTENT)


class UpdateInvoiceAddressCartView(generics.CreateAPIView):
    serializer_class = serializers.InvoiceAddressCartSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        cart = models.Cart(
            cart_token=request.session['cart_token'],
            user=request.user
        )
        cart.update_invoice_address(**serializer.data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class UpdateShippingAddressCartView(generics.CreateAPIView):
    serializer_class = serializers.ShippingAddressCartSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        cart = models.Cart(
            cart_token=request.session['cart_token'],
            user=request.user
        )
        cart.update_shipping_address(**serializer.data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class UpdateNoteCartView(generics.CreateAPIView):
    serializer_class = serializers.NoteCartSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        cart = models.Cart(
            cart_token=request.session['cart_token'],
            user=request.user
        )
        cart.update_note(**serializer.data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class RemoveCartView(generics.DestroyAPIView):
    serializer_class = serializers.CartSerializer

    def get_object(self):
        cart = models.Cart(
            cart_token=self.request.session['cart_token'],
            user=self.request.user
        )
        cart.delete()


class OrderCartView(generics.RetrieveAPIView):
    serializer_class = serializers.OrderSerializer

    def get_owner(self):
        owners = Organization.objects.filter(user=self.request.user)
        if owners:
            return owners[0]
        return Organization.objects.none()

    def retrieve(self, request, *args, **kwargs):
        user = self.request.user
        if not user.is_anonymous():
            cart = models.Cart(
                cart_token=self.request.session['cart_token'],
                user=self.request.user
            )
            order = cart.create_order()
            if order:
                serializer = self.get_serializer(order)
                return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response({})


class OrderListView(generics.ListAPIView):
    serializer_class = serializers.OrderSerializer

    def get_queryset(self):
        user = self.request.user
        if user.is_anonymous():
            return models.Order.objects.none()
        else:
            return models.Order.objects.filter(owner__user=user)


class OrderDetailView(generics.RetrieveAPIView):
    serializer_class = serializers.OrderSerializer

    def get_queryset(self):
        user = self.request.user
        if user.is_anonymous():
            return models.Order.objects.none()
        else:
            return models.Order.objects.filter(owner__user=user)


class OrderRowListView(generics.ListAPIView):
    serializer_class = serializers.OrderRowSerializer

    def get_queryset(self):
        user = self.request.user
        if user.is_anonymous():
            return models.OrderRow.objects.none()
        else:
            return models.OrderRow.objects.filter(
                order__owner__user=user, order_id=self.kwargs['order_id']
            )
