# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from six import python_2_unicode_compatible
import calendar
import datetime
import decimal
import string

from django.conf import settings
from django.contrib.auth.models import User
from django.core.mail import send_mail
from django.core.urlresolvers import reverse
from django.db import models
from django.db.models import Q
from django.template import Context
from django.template.loader import render_to_string
from django.utils.crypto import get_random_string
from django.utils.translation import ugettext_lazy as _

from jmb.core.db.models import DateModel, UserModel, StatusModel, OrderedModel, DefaultModel

from thux.bank.models import Bank
from thux.carriage.models import Delivery
from thux.currency.models import Currency
from thux.l10n.models import Country
from thux.office.models import Office
from thux.organization.models import Organization, Contact, OrganizationAddress
from thux.payment.models import PaymentForm
from thux.product.models import Product, UnitType
from thux.store.models import StoreCause
from thux.vat.models import Vat


VALID_KEY_CHARS = string.ascii_lowercase + string.digits


@python_2_unicode_compatible
class Cart(object):
    @classmethod
    def merge_carts(
        cls, user_cart, token_cart, user, cart_token,
        cart_invoice_address=None, cart_shipping_address=None, cart_note=None
    ):
        if user_cart != token_cart:
            user_rows = user_cart.cart_rows
            token_rows = token_cart.cart_rows

            for user_row in user_rows:
                try:
                    token_row = token_rows.get(product=user_row.product)
                    if token_row.quantity > user_row.quantity:
                        user_row.quantity = token_row.quantity
                        user_row.save()
                    token_row.delete()
                except:
                    pass
            token_rows.update(user=user)

        if cart_invoice_address:
            cart_invoice_address(user=user, cart_token=cart_token)

        if cart_shipping_address:
            cart_shipping_address(user=user, cart_token=cart_token)

        if cart_note:
            cart_note(user=user, cart_token=cart_token)

        return Cart(cart_token, user)

    @classmethod
    def create_cart_token(cls):
        """
        Return cart token that isn't being used.
        """
        while True:
            cart_token = get_random_string(32, VALID_KEY_CHARS)
            if not cls.exists_cart_token(cart_token):
                break
        return cart_token

    @classmethod
    def exists_cart_token(cls, cart_token):
        return CartRow.objects.filter(cart_token=cart_token).exists()

    def __str__(self):
        return  self.cart_token

    def __init__(self, cart_token, user=None):
        self.cart_token = cart_token
        self.user = None
        if user and user.is_authenticated():
            self.user = user

    def __eq__(self, other):
        return self.cart_token == other.cart_token and self.user == other.user

    def get_cart_rows(self):
        if self.user:
            # prende tutti i carelli attivi dello user o della sessione impostati
            return CartRow.objects.filter(
                Q(user=self.user) | Q(cart_token=self.cart_token)
            )
        else:
            # prende il carrello di quella sessione
            return CartRow.objects.filter(
                cart_token=self.cart_token,
                user__isnull=True
            )
    cart_rows = property(get_cart_rows)

    def add_product(self, product, quantity=1):
        try:
            row = self.cart_rows.get(product=product)
            row.quantity += quantity
        except CartRow.DoesNotExist:
            row = CartRow(
                cart_token=self.cart_token,
                user=self.user,
                product=product,
                quantity=quantity,
            )
        row.save()
        return row

    def update_product(self, product, quantity=1):
        try:
            row = self.cart_rows.get(product=product)
            row.quantity = quantity
        except:
            row = CartRow(
                cart_token=self.cart_token,
                user=self.user,
                product=product,
                quantity=quantity,
            )
        row.save()
        return row

    def delete_product(self, product_pk):
        try:
            row = self.cart_rows.get(product=product_pk)
            row.delete()
        except:
            pass

    def get_shipping_fee(self):
        """
        Calculate the shipping fee
        """
        if 0 < self.partial_total < settings.ORDER_CART_SHIPPING_FEE_THRESHOLD:
            return settings.ORDER_CART_SHIPPING_FEE_COST
        return 0
    shipping_fee = property(get_shipping_fee)

    def get_partial_total(self):
        """
        Total without shipping fee
        """
        rows = self.cart_rows
        total = 0
        for row in rows:
            total += row.total
        return total
    partial_total = property(get_partial_total)

    def get_total(self):
        """
        Calculate total with shipping fee
        """
        part_total = float(self.partial_total)
        fee = self.shipping_fee
        return part_total + fee
    total = property(get_total)

    def get_count(self):
        rows = self.cart_rows
        count = 0
        for row in rows:
            count += row.quantity
        return count
    count = property(get_count)

    def get_invoice_address(self):
        try:
            if self.user:
                return CartInvoiceAddress.objects.get(
                    user=self.user
                )
            else:
                return CartInvoiceAddress.objects.get(
                    cart_token=self.cart_token,
                    user__isnull=True
                )
        except:
            return CartInvoiceAddress.objects.none()
    invoice_address = property(get_invoice_address)

    def update_invoice_address(self, **kwargs):
        try:
            address = self.invoice_address
            address = address(
                address_type=kwargs['address_type'], name=kwargs['name'],
                number=kwargs['number'], crossed=kwargs['crossed'],
                scale=kwargs['scale'], floor=kwargs['floor'],
                zip_code=kwargs['zip_code'], locality=kwargs['locality'],
                province=kwargs['province'], country_id=kwargs['country'],
                post_office=kwargs['post_office'],
            )
        except:
            address = CartInvoiceAddress(
                cart_token=self.cart_token, user=self.user,
                address_type=kwargs['address_type'], name=kwargs['name'],
                number=kwargs['number'], crossed=kwargs['crossed'],
                scale=kwargs['scale'], floor=kwargs['floor'],
                zip_code=kwargs['zip_code'], locality=kwargs['locality'],
                province=kwargs['province'], country_id=kwargs['country'],
                post_office=kwargs['post_office'],
            )
        address.save()
        return address

    def get_shipping_address(self):
        try:
            if self.user:
                return CartShippingAddress.objects.get(
                    user=self.user
                )
            else:
                return CartShippingAddress.objects.get(
                    cart_token=self.cart_token,
                    user__isnull=True
                )
        except:
            return CartShippingAddress.objects.none()
    shipping_address = property(get_shipping_address)

    def update_shipping_address(self, **kwargs):
        try:
            address = self.invoice_address
            address = address(
                address_type=kwargs['address_type'], name=kwargs['name'],
                number=kwargs['number'], crossed=kwargs['crossed'],
                scale=kwargs['scale'], floor=kwargs['floor'],
                zip_code=kwargs['zip_code'], locality=kwargs['locality'],
                province=kwargs['province'], country_id=kwargs['country'],
                post_office=kwargs['post_office'],
            )
        except:
            address = CartShippingAddress(
                cart_token=self.cart_token, user=self.user,
                address_type=kwargs['address_type'], name=kwargs['name'],
                number=kwargs['number'], crossed=kwargs['crossed'],
                scale=kwargs['scale'], floor=kwargs['floor'],
                zip_code=kwargs['zip_code'], locality=kwargs['locality'],
                province=kwargs['province'], country_id=kwargs['country'],
                post_office=kwargs['post_office'],
            )
        address.save()
        return address

    def get_note(self):
        try:
            if self.user:
                return CartNote.objects.get(
                    user=self.user
                )
            else:
                return CartNote.objects.get(
                    cart_token=self.cart_token,
                    user__isnull=True
                )
        except:
            return CartNote.objects.none()
    note = property(get_note)

    def update_note(self, **kwargs):
        try:
            note = self.note
            note = note(note=kwargs['note'])
        except:
            note = CartNote(
                cart_token=self.cart_token, user=self.user,
                note=kwargs['note']
            )
        note.save()
        return note

    def delete(self):
        for row in self.cart_rows:
            row.delete()
        if self.invoice_address:
            self.invoice_address.delete()
        if self.shipping_address:
            self.shipping_address.delete()
        if self.note:
            self.note.delete()

    def create_order(self, paypal_payment_id):
        shipping_address = CartShippingAddress.objects.get(cart_token=self.cart_token)
        try:
            invoice_address = CartInvoiceAddress.objects.get(cart_token=self.cart_token)
        except CartInvoiceAddress.ObjectDoesNotExist:
            invoice_address = None
        try:
            note = CartNote.objects.get(cart_token=self.cart_token).note
        except CartInvoiceAddress.ObjectDoesNotExist:
            note = ''

        if not invoice_address:
            organization = Organization.objects.get(user=shipping_address.user)

            invoice_address = CartInvoiceAddress(
                cart_token=shipping_address.cart_token,
                nominative=organization.name,
                ssn=shipping_address.ssn,
                name=shipping_address.name,
                zip_code=shipping_address.zip_code,
                locality=shipping_address.locality,
                province=shipping_address.province,
                country_id=shipping_address.country.id,
            )
            invoice_address.save()

        if self.cart_rows:
            order = Order(
                status='ord',
                owner=self.user.organization,
                # owner_number=owner_number,
                note=note,
                # TODO
                total_income=self.total,

                # INVOICE ADDRESS
                invoice_name=invoice_address.name,
                invoice_zip_code=invoice_address.zip_code,
                invoice_locality=invoice_address.locality,
                invoice_province=invoice_address.province,
                invoice_country_id=invoice_address.country.id,

                # SHIPPING ADDRESS
                shipping_name=shipping_address.name,
                shipping_zip_code=shipping_address.zip_code,
                shipping_locality=shipping_address.locality,
                shipping_province=shipping_address.province,
                shipping_country_id=shipping_address.country.id
            )
            order.save()

            for row in self.cart_rows:
                ord_row = OrderRow(
                    order=order, product=row.product,
                    quantity=row.quantity, price=row.product.street_price
                )
                ord_row.save()

                # REMOVE CART_ROW
                row.delete()

            # REMOVE CART
            self.delete()
            return order
        return None

    def get_paypal_items(self):
        items = []
        for cart_row in self.cart_rows:
            item = {
                'name': cart_row.product.name,
                'sku': cart_row.product.pk,
                'price': float(cart_row.product.get_price_vat_included()),
                'currency': 'EUR',
                'quantity': int(cart_row.quantity),
            }
            items.append(item)
        # appending shipping fees
        fees = {
            'name': 'Shipping Fees',
            'sku': 0,
            'price': float(self.shipping_fee),
            'currency': 'EUR',
            'quantity': 1,
        }
        items.append(fees)
        return items

    def pay_with_paypal(self):
        """
        Method to launch paypal payment, to complete the payment is necessary navigate paypal checkout url:
        self.paypal.approval_link
        :return: payment structure dict
        """
        items = self.get_paypal_items()
        self.paypal.pay_with_paypal(
            init=True,
            items=items,
            return_url=settings.BASE_URL + reverse(settings.PAYPAL_RETURN_URL_NAME),
            cancel_url=settings.BASE_URL + reverse(settings.PAYPAL_CANCEL_URL_NAME),
        )
        return self.paypal.payment


@python_2_unicode_compatible
class CartInvoiceAddress(DateModel, UserModel):
    cart_token = models.CharField(
        max_length=100,
        verbose_name=_('session key'),
    )
    nominative = models.CharField(
        max_length=255,
        verbose_name=_('nominative'),
    )
    ssn = models.CharField(
        max_length=16,
        verbose_name=_('social security number'),
    )
    user = models.ForeignKey(
        User,
        blank=True, null=True,
        related_name='cart_invoice_addresses',
        verbose_name=_('user'),
    )
    address_type = models.IntegerField(
        blank=True, null=True,
        choices=settings.ADDRESS_ADDRESS_TYPE_CHOICES,
        db_index=True,
        verbose_name=_("type"),
    )
    name = models.CharField(
        max_length=100,
        blank=True,
        db_index=True,
        verbose_name=_("name"),
    )
    number = models.CharField(
        max_length=20,
        blank=True,
        verbose_name=_("number"),
        db_column="inv_number"
    )
    crossed = models.CharField(
        max_length=5,
        blank=True,
        verbose_name=_("crossed"),
    )
    scale = models.CharField(
        max_length=5,
        blank=True,
        verbose_name=_("scale"),
    )
    floor = models.CharField(
        max_length=3,
        blank=True,
        verbose_name=_("floor"),
    )
    zip_code = models.CharField(
        max_length=20,
        db_index=True,
        blank=True,
        verbose_name=_("zip code"),
    )
    locality = models.CharField(
        max_length=100,
        db_index=True,
        blank=True,
        verbose_name=_("locality"),
    )
    province = models.CharField(
        max_length=2,
        db_index=True,
        blank=True,
        verbose_name=_("province"),
    )
    country = models.ForeignKey(
        Country,
        related_name='cart_invoice_addresses',
        blank=True, null=True,
        verbose_name=_("country"),
    )
    post_office = models.CharField(
        max_length=10,
        blank=True,
        verbose_name=_("post office"),
    )

    def __str__(self):
        return "{} - {}: {}".format(
            self.cart_token, self.user,
            self.name
        )


@python_2_unicode_compatible
class CartShippingAddress(DateModel, UserModel):
    cart_token = models.CharField(
        max_length=100,
        verbose_name=_('session key'),
    )
    first_name = models.CharField(
        max_length=255,
        verbose_name=_('first name'),
    )
    last_name = models.CharField(
        max_length=255,
        verbose_name=_('last name'),
    )
    ssn = models.CharField(
        max_length=16,
        verbose_name=_('social security number'),
    )
    user = models.ForeignKey(
        User,
        blank=True, null=True,
        related_name='cart_shipping_addresses',
        verbose_name=_('user'),
    )
    address_type = models.IntegerField(
        blank=True, null=True,
        choices=settings.ADDRESS_ADDRESS_TYPE_CHOICES,
        db_index=True,
        verbose_name=_("type"),
    )
    name = models.CharField(
        max_length=100,
        blank=True,
        db_index=True,
        verbose_name=_("name"),
    )
    number = models.CharField(
        max_length=20,
        blank=True,
        verbose_name=_("number"),
        db_column="inv_number"
    )
    crossed = models.CharField(
        max_length=5,
        blank=True,
        verbose_name=_("crossed"),
    )
    scale = models.CharField(
        max_length=5,
        blank=True,
        verbose_name=_("scale"),
    )
    floor = models.CharField(
        max_length=3,
        blank=True,
        verbose_name=_("floor"),
    )
    zip_code = models.CharField(
        max_length=20,
        db_index=True,
        blank=True,
        verbose_name=_("zip code"),
    )
    locality = models.CharField(
        max_length=100,
        db_index=True,
        blank=True,
        verbose_name=_("locality"),
    )
    province = models.CharField(
        max_length=2,
        db_index=True,
        blank=True,
        verbose_name=_("province"),
    )
    country = models.ForeignKey(
        Country,
        related_name='cart_shipping_addresses',
        blank=True, null=True,
        verbose_name=_("country"),
    )
    post_office = models.CharField(
        max_length=10,
        blank=True,
        verbose_name=_("post office"),
    )

    def __str__(self):
        return "{} - {}: {}".format(
            self.cart_token, self.user,
            self.name
        )


@python_2_unicode_compatible
class CartNote(DateModel, UserModel):
    cart_token = models.CharField(
        max_length=100,
        verbose_name=_('session key'),
    )
    user = models.ForeignKey(
        User,
        blank=True, null=True,
        related_name='cart_notes',
        verbose_name=_('user'),
    )
    note = models.TextField(
        blank=True,
        verbose_name=_("note"),
    )

    class Meta:
        verbose_name = _("cart note")
        verbose_name_plural = _("cart notes")

    def __str__(self):
        return "{} - {}: {}".format(
            self.cart_token, self.user,
            self.note
        )


@python_2_unicode_compatible
class CartRow(DateModel, UserModel):
    @classmethod
    def merge_cart_rows(cls, user, cart_token):
        if not user.is_anonymous():
            user_rows = cls.objects.filter(user=user)
            token_rows = cls.objects.filter(cart_token=cart_token, user__isnull=True)

            # mergia il carrello (vecchio) dell'utente loggato con il carrello (nuovo) dell'utente non loggato
            if token_rows:
                for user_row in user_rows:
                    try:
                        token_row = token_rows.get(product=user_row.product)
                        if token_row.quantity > user_row.quantity:
                            user_row.quantity = token_row.quantity
                            user_row.save()
                        token_row.delete()
                    except:
                        pass
                token_rows.update(user=user)

    cart_token = models.CharField(
        max_length=100,
        verbose_name=_('session key'),
    )
    user = models.ForeignKey(
        User,
        blank=True, null=True,
        related_name='cart_rows',
        verbose_name=_('user'),
    )
    product = models.ForeignKey(
        Product,
        related_name='cart_rows',
        verbose_name=_('product'),
    )
    quantity = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        verbose_name=_('quantity'),
    )

    def get_total(self):
        return self.product.get_price_vat_included() * self.quantity
    total = property(get_total)

    class Meta:
        verbose_name = _("cart row")
        verbose_name_plural = _("cart rows")
        unique_together = ('cart_token', 'user', 'product')
        ordering = ("pk",)

    def __str__(self):
        return "{} - {}: {} {}".format(
            self.cart_token, self.user,
            self.quantity, self.product.name
        )

    def get_cart(self):
        return Cart(self.cart_token, user=self.user)
    cart = property(get_cart)


@python_2_unicode_compatible
class OrderNumber(DateModel, UserModel):
    @classmethod
    def set_new_number(cls, year=None):
        year = year or datetime.datetime.today().year
        order_numbers = cls.objects.filter(year=year)
        if order_numbers:
            order_number = order_numbers[0]
            order_number.last_number += 1
        else:
            order_number = cls()
            order_number.year = year
            order_number.last_number = 1
        order_number.save()
        return order_number

    year = models.PositiveIntegerField(
        unique=True,
        verbose_name=_('year'),
    )
    last_number = models.PositiveIntegerField(
        verbose_name=_('last number'),
    )

    class Meta:
        verbose_name = _("order number")
        verbose_name_plural = _("order numbers")
        ordering = ('-year', '-last_number',)
        permissions = (
            ("list_ordernumber", "Can list order number"),
        )

    def __str__(self):
        return "{}: {}".format(self.year, self.last_number)


@python_2_unicode_compatible
class Order(DateModel, UserModel):
    status = models.CharField(
        choices=settings.ORDER_ORDER_STATUS,
        max_length=3, default=0,
        db_index=True,
        verbose_name=_('status'),
    )
    owner = models.ForeignKey(
        Organization,
        related_name='owner_orders',
        verbose_name=_('bill to'),
    )
    recipient = models.ForeignKey(
        Organization,
        null=True, blank=True,
        related_name='recipient_orders',
        verbose_name=_('Ship to'),
    )
    # project = models.ForeignKey(
    # Project,
    # blank=True, null=True,
    # related_name='orders',
    # verbose_name=_('project'),
    # )
    payment_form = models.ForeignKey(
        PaymentForm,
        blank=True, null=True,
        related_name='orders',
        verbose_name=_('payment form'),
    )
    office = models.ForeignKey(
        Office,
        blank=True, null=True,
        related_name='orders',
        verbose_name=_('office'),
    )
    contact = models.ForeignKey(
        Contact,
        blank=True, null=True,
        related_name='orders',
        verbose_name=_('contact'),
    )
    bank = models.ForeignKey(
        Bank,
        blank=True, null=True,
        related_name='orders',
        verbose_name=_('bank')
    )
    currency = models.ForeignKey(
        Currency,
        related_name='orders',
        verbose_name=_('currency')
    )
    delivery = models.ForeignKey(
        Delivery,
        blank=True, null=True,
        related_name='orders',
        verbose_name=_('delivery')
    )
    vat = models.ForeignKey(
        Vat,
        blank=True, null=True,
        related_name='orders',
        verbose_name=_('vat')
    )
    # TODO: occorre creare modello proxy
    # agent = models.ForeignKey(
    #     Agent,
    #     null = True, blank = True,
    #     related_name = 'orders',
    #     verbose_name = _('agent'),
    # )
    # TODO: occorre creare modello
    # commision = models.ForeignKey (
    #     Commision,
    #     blank = True, null = True,
    #     related_name = 'orders',
    #     verbose_name = _('agent & commision')
    # )
    # price_list = models.ForeignKey(
    #     OrganizationPriceList,
    #     blank=True, null=True,
    #     related_name='orders',
    #     verbose_name=_('price list & discount')
    # )
    # TODO: occorre creare modello proxy
    # vector = models.ForeignKey(
    #     Vector,
    #     blank = True, null = True,
    #     related_name = 'orders',
    #     verbose_name = _('vector')
    # )
    store_cause = models.ForeignKey(
        StoreCause,
        related_name='orders',
        verbose_name=_('Store Cause')
    )
    shipping_address = models.ForeignKey(
        OrganizationAddress,
        blank=True, null=True,
        related_name='orders',
        verbose_name=_('shipping address')
    )
    description = models.CharField(
        max_length=255,
        db_index=True,
        verbose_name=_('description')
    )
    number = models.CharField(
        max_length=50,
        blank=True,
        db_index=True,
        verbose_name=_('number'),
        db_column='pnumber'
    )
    owner_number = models.CharField(
        max_length=50,
        blank=True,
        db_index=True,
        verbose_name=_('ref owner number'),
    )
    date_p = models.DateField(
        db_index=True,
        verbose_name=_('date')
    )
    ddt_completed = models.BooleanField(
        default=False,
        db_index=True,
        verbose_name=_('DDT completed'),
    )
    invoice_completed = models.BooleanField(
        default=False,
        db_index=True,
        verbose_name=_('invoice completed'),
    )
    note = models.TextField(
        blank=True,
        verbose_name=_('note'),
    )
    internal_note = models.TextField(
        blank=True,
        verbose_name=_('internal note'),
    )
    accepted = models.BooleanField(
        default=False,
        db_index=True,
        verbose_name=_('accepted'),
    )
    total_income = models.DecimalField(
        default=0,
        max_digits=11,
        decimal_places=4,
        verbose_name=_('total income'),
    )
    total_fee = models.DecimalField(
        default=0,
        max_digits=11,
        decimal_places=4,
        verbose_name=_('total fee'),
    )
    total_discount = models.DecimalField(
        default=0,
        max_digits=11,
        decimal_places=4,
        verbose_name=_('total discount'),
    )
    total_vat = models.DecimalField(
        default=0,
        max_digits=11,
        decimal_places=4,
        verbose_name=_('total VAT'),
    )
    total = models.DecimalField(
        default=0,
        max_digits=12,
        decimal_places=4,
        verbose_name=_('total'),
    )
    other_shipping_address_owner = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('other shipping address owner'),
    )
    other_shipping_address = models.CharField(
        max_length=255,
        blank=True, null=True,
        verbose_name=_('other shipping address'),
    )
    date_delivery = models.DateField(
        blank=True, null=True,
        db_index=True,
        verbose_name=_('date delivery'),
    )
    ## campi condizioni generali
    # validita offerta
    quote_firm_for = models.CharField(
        max_length=255,
        blank=True, null=True,
        verbose_name=_('quote firm for'),
    )
    # tempi di consegna
    delivery_window = models.CharField(
        max_length=255,
        blank=True, null=True,
        verbose_name=_('delivery window'),
    )

    class Meta:
        verbose_name = _('order')
        verbose_name_plural = _('orders')
        ordering = ('-date_p', '-number')
        permissions = (
            ("list_order", "Can list order"),
        )

    def __str__(self):
        return '{}, {} - total: {}'.format(
            self.date_p, self.number, self.total
        )

    def save(self, *args, **kwargs):
        if not self.number:
            # TODO:
            ## solo se l'ordine è nuovo e non ha ancora un numero assegnato.
            self.number = OrderNumber.set_new_number(
                year=self.ord_date.year
            ).last_number

        if self.pk:
            old_obj = Order.objects.get(pk=self.pk)

            old_accepted = old_obj.accepted
            old_invoice = old_obj.invoice
            old_ddt = old_obj.ddt

            change = False

            if old_accepted != self.accepted:
                change = True
            if old_ddt != self.ddt:
                change = True
            if old_invoice != self.invoice:
                change = True

            if change and settings.ORDER_SEND_ORDER_PROCESS_BY_MAIL:
                self.send_mail_order_process_status()
        super(Order, self).save(*args, **kwargs)

    def store_rows(self):
        return OrderRow.objects.filter(
            order=self, ddt=False
        ).order_by('product__positions__position').distinct()

    def get_payment_installments(self):
        if not self.payment_form:
            return []

        if not self.payment_form.installments.all():
            return []

        installments = []
        total_amount = 0
        for inst in self.payment_form.installments.all():
            installment = {
                'name': inst.name, 'full_name': inst.full_name,
                'maturity': self.date_p + datetime.timedelta(days=inst.days)}
            if inst.end_month:
                maturity = installment['maturity']
                monthrange = calendar.monthrange(
                    maturity.year,
                    maturity.month
                )
                installment['maturity'] = (
                    datetime.date(maturity.year, maturity.month, monthrange[1]) +
                    datetime.timedelta(days=inst.days_end_month)
                )
            ## Arrotondo la divisione e lascio il resto per l'ultima rata
            installment['amount'] = int(
                self.totals['total'] * inst.amount / 100
            )
            ## controllo l'arrotondamento
            total_amount += installment['amount']
            installments.append(installment)
        ## controllo l'arrotondamento
        if total_amount != self.totals['total']:
            installments[-1]['amount'] = installments[-1]['amount'] - total_amount + self.totals['total']
        return installments

    def default_invoice_address(self):
        invoice_addresses = self.owner.get_addresses().filter(address_scopes__is_invoice_address=True)
        if invoice_addresses:
            return invoice_addresses[0]
        return None

    def default_invoice_address1(self):
        addr = self.default_invoice_address()
        return '{} {}, {}'.format(addr.get_address_type_display() or '', addr.name, addr.number)

    def default_invoice_address2(self):
        addr = self.default_invoice_address()
        return '{} {} ({})'.format(addr.zip_code, addr.locality, addr.province)

    def default_invoice_address3(self):
        addr = self.default_invoice_address()
        return addr.country

    def shipping_address1(self):
        return '{} {}, {}'.format(
            self.shipping_address.get_address_type_display() or '',
            self.shipping_address.name, self.shipping_address.number
        )

    def shipping_address2(self):
        return '{} {} ({})'.format(
            self.shipping_address.zip_code, self.shipping_address.locality,
            self.shipping_address.province
        )

    def shipping_address3(self):
        return self.shipping_address.country

    def update_totals(self):
        try:
            first_orderrow = self.rows.all()[0]
            first_orderrow.update_totals()
        except Exception:
            pass

    def get_deliveredquantity(self):
        qty_delivered = 0
        for row in self.rows.all():
            qty_delivered += row.get_deliveredquantity()
        return qty_delivered

    def get_notdeliveredquantity(self):
        qty_not_delivered = 0
        for row in self.rows.all():
            qty_not_delivered += row.get_notdeliveredquantity()
        return qty_not_delivered

    def get_deliveredvalue(self):
        if self.ddt:
            return self.total_income

        qty_delivered = 0
        for row in self.rows.all():
            qty_delivered += row.get_deliveredvalue()
        return qty_delivered

    def get_notdeliveredvalue(self):
        return self.total_income - self.get_deliveredvalue()

    def get_backorder_quantity_and_value(self):
        qty_not_delivered = 0
        for row in self.rows:
            qty_not_delivered += row.get_notdeliveredquantity
        return qty_not_delivered

    def get_totals(self):
        # Ho dovuto mettere la forzatura a Decimal solamente perche'
        # questo metodo viene chiamato anche nella funzione __unicode__.
        # Non so perche' ma quando viene chiamato da quella specifica funzione
        # i valori self.totale_income, self.total_fee, etc sono delle stringhe,
        # mentre normalemente (e giustamente) sono dei decimal.
        #
        totals = dict()
        totals['income'] = decimal.Decimal(self.total_income)
        totals['fee'] = decimal.Decimal(self.total_fee)
        totals['discount'] = decimal.Decimal(self.total_discount)
        totals['vat'] = decimal.Decimal(self.total_vat)
        totals['total_no_vat'] = (
            decimal.Decimal(self.total_income) +
            decimal.Decimal(self.total_fee) -
            decimal.Decimal(self.total_discount)
        )
        totals['total'] = totals['total_no_vat'] + decimal.Decimal(self.total_vat)
        return totals

    totals = property(get_totals)

    def send_message_order_added(self, to_email=None):
        """
        È l'email inviata quando viene inserito un ordine
        """
        url = settings.BASE_URL + reverse('admin:order_order_change', args=[self.id])
        from_mail = settings.NO_REPLY_EMAIL
        if not to_email:
            to_email = settings.ADMIN_EMAIL
        context = Context({
            'order': self,
            'url': url,
        })
        subject = _("Order no. {} of {} added").format(self.number, self.ord_date)

        # messaggio testuale
        text_message = render_to_string(
            'order/order/email/send_message_order_added.txt',
            context,
        )

        # messaggio html
        html_message = render_to_string(
            'order/order/email/send_message_order_added.html',
            context,
        )
        try:
            msg = send_mail(
                subject=subject,
                message=text_message,
                html_message=html_message,
                recipient_list=[to_email],
                from_email=from_mail,
            )
            return True, msg
        except Exception as e:
            return False, e

    def send_mail_order_process_status(self, to_email=None):
        pass


@python_2_unicode_compatible
class OrderRow(DateModel, UserModel, StatusModel, OrderedModel):
    order = models.ForeignKey(
        Order,
        related_name='order_rows',
        verbose_name=_('order'),
    )
    product = models.ForeignKey(
        Product,
        related_name='order_rows',
        verbose_name=_('product'),
    )
    vat = models.ForeignKey(
        Vat,
        related_name='order_rows',
        verbose_name=_('Vat'),
    )
    store_cause = models.ForeignKey(
        StoreCause,
        related_name='order_rows',
        verbose_name=_('Store Cause'),
    )
    # project = models.ForeignKey(
    #     Project,
    #     null=True, blank=True,
    #     related_name='order_rows',
    #     verbose_name=_('Project'),
    # )
    quantity = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        default=0,
        verbose_name=_('quantity'),
    )
    unit = models.ForeignKey(
        UnitType,
        default='pz',
        db_index=True,
        verbose_name=_('unit'),
    )
    price = models.DecimalField(
        max_digits=11,
        decimal_places=4,
        default=0,
        verbose_name=_('unit price'),
    )
    not_apply_discount = models.BooleanField(
        default=False,
        verbose_name=_('no discount')
    )
    fee = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        default=0,
        verbose_name=_('fee'),
    )
    discount = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        default=0,
        verbose_name=_('discount'),
    )
    perc_discount1 = models.DecimalField(
        max_digits=5,
        decimal_places=2,
        default=0,
        verbose_name=_('discount 1 %')
    )
    perc_discount2 = models.DecimalField(
        max_digits=5,
        decimal_places=2,
        default=0,
        verbose_name=_('discount 2 %'),
    )
    ddt_completed = models.BooleanField(
        db_index=True,
        verbose_name=_('DDT completed'),
    )
    invoice_completed = models.BooleanField(
        db_index=True,
        verbose_name=_('invoice completed'),
    )
    description = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_('description')
    )
    long_description = models.TextField(
        blank=True,
        verbose_name=_('long description'),
    )
    note = models.TextField(
        blank=True,
        verbose_name=_('note'),
    )
    to_print = models.BooleanField(
        db_index=True,
        default=True,
        verbose_name=_('to print'),
    )
    # data prevista consegna
    date_delivery = models.DateField(
        blank=True, null=True,
        db_index=True,
        verbose_name=_('date delivery'),
    )

    class Meta:
        verbose_name = _('order row')
        verbose_name_plural = _('order rows')
        ordering = ('-order__date_p', '-order__number', 'ordering', 'id')
        permissions = (
            ("list_orderrow", "Can list order row"),
        )

    def __str__(self):
        return '{}, {}'.format(self.order, self.product)

    def save(self, *args, **kwargs):
        super(OrderRow, self).save(*args, **kwargs)
        self.update_totals()

    def save_not_update(self,  *args, **kwargs):
        super(OrderRow, self).save( *args, **kwargs)

    def delete(self, *args, **kwargs):
        super(OrderRow, self).delete(*args, **kwargs)
        self.update_totals()

    def get_description(self):
        """
        Restituisce la descrizione della riga o, se non esiste, quella del prodotto
        """
        return self.description or self.product.description

    def get_list_positions(self):
        """
        Resituisce la lista delle posizioni del prodotto nel magazzino selezionato
        """
        product_list = []
        for position in self.product.positions.filter(store=self.store_cause.store):
            product_list.append('{}'.append(position.position))
        return product_list

    def get_net_price(self):
        # TODO: da verificare
        price = decimal.Decimal(self.price)

        # Se NOT_APPLY_DISCOUNT e' flaggato non calcolo gli sconti
        if not self.not_apply_discount and self.order.price_list:
            d1 = self.order.price_list.discount1
            d2 = self.order.price_list.discount2
            price = (self.price + self.fee - self.discount)  # * (1 - d1/100) * (1 - d2/100)
            ## arrotondamento per difetto e per eccesso
            price += decimal.Decimal('0.00005')
        return price
    get_net_price.short_description = _('net price')
    net_price = property(get_net_price)

    def get_total_income(self):
        """
        Restituisce il totale imponibile calcolato sulla Riga: quantita x prezzo
        """
        return self.quantity * self.price
    get_total_income.short_description = _('total income')
    total_income = property(get_total_income)

    def get_total_fee(self):
        """
        Restituisce il totale spese calcolato sulla Riga: quantita x spesa accessoria
        """
        return self.quantity * self.fee
    get_total_fee.short_description = _('total fee')
    total_fee = property(get_total_fee)

    def get_total_discount(self):
        """
        Restituisce il totale sconto calcolato sulla Riga: quantita x sconto
        """
        # Nel calcolo dello sconto devo arrotondare tt a 2 decimali altrimenti
        # i totali dell'ordine e dei ddt sballano di qualche decimale
        #
        discount1 = self.price * self.perc_discount1 / 100
        discount2 = (self.price - discount1) * self.perc_discount2 / 100
        unit_discount = self.discount + discount1 + discount2
        return self.quantity * unit_discount
    get_total_discount.short_description = _('total discount')
    total_discount = property(get_total_discount)

    def get_total(self):
        """
        Restituisce il Totale generale calcolato sommando e sottrando tutti i totali
        """
        return self.total_income + self.total_fee - self.total_discount
    get_total.short_description = _('total')
    total = property(get_total)

    def update_totals(self):
        """
        Aggiorna i totali dell'ordine e delle righe relative all'IVA
        """
        order = self.order
        order_rows = self.order.order_rows.all()
        VatRow.objects.filter(order=order).delete()
        total = {'income': 0.0000, 'fee': 0.0000, 'discount': 0.0000, 'total': 0.0000, 'vat': 0.0000}
        subtotal = {'income': {}, 'fee': {}, 'discount': {}, 'vat': {}, 'total': {}}

        for row in order_rows:
            subtotal['income'][row.vat] = subtotal['income'].get(row.vat, None) or 0.0000
            subtotal['income'][row.vat] += row.total_income

            subtotal['fee'][row.vat] = subtotal['fee'].get(row.vat, None) or 0.0000
            subtotal['fee'][row.vat] += row.total_fee

            subtotal['discount'][row.vat] = subtotal['discount'].get(row.vat, None) or 0.0000
            subtotal['discount'][row.vat] += row.total_discount

            subtotal['total'][row.vat] = subtotal['total'].get(row.vat, None) or 0.0000
            subtotal['total'][row.vat] += row.total

        for vat in subtotal['income'].keys():
            ## calcolo Totali Procedura
            total['income'] += subtotal['income'][vat]
            total['fee'] += subtotal['fee'][vat]
            total['discount'] += subtotal['discount'][vat]
            total['total'] += subtotal['total'][vat]
            ## calcolo Subtotale IVA per ogni aliquota
            subtotal['vat'][vat] = subtotal['total'][vat] * vat.aliquot_perc / 100.0000
            total['vat'] += subtotal['vat'][vat]
            ## creo nuovo oggetto VatRow
            vatrow = VatRow()
            vatrow.order = order
            vatrow.vat = vat
            vatrow.total_income = subtotal['income'][vat]
            vatrow.total_fee = subtotal['fee'][vat]
            vatrow.total_discount = subtotal['discount'][vat]
            vatrow.total_vat = subtotal['vat'][vat]
            vatrow.save()

        total['totals'] = total['total'] + total['vat']

        order.total_income = total['income']
        order.total_fee = total['fee']
        order.total_discount = total['discount']
        order.total_vat = total['vat']
        order.save()


@python_2_unicode_compatible
class VatRow (DateModel, UserModel, StatusModel):
    order = models.ForeignKey(
        Order,
        related_name='vat_total_rows',
        verbose_name=_('order'),
    )
    vat = models.ForeignKey(
        Vat,
        related_name='vat_total_rows',
        verbose_name=_('vat'),
    )
    total_income = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        verbose_name=_('total income'),
    )
    total_fee = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        verbose_name=_('total fee'),
    )
    total_discount = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        verbose_name=_('total discount'),
    )
    total_vat = models.DecimalField(
        max_digits=9,
        decimal_places=2,
        verbose_name=_('total vat'),
    )

    class Meta:
        verbose_name = _('vat row')
        verbose_name_plural = _('vat rows')
        ordering = ('vat__ordering', 'vat__code')
        unique_together = (('order', 'vat'),)
        permissions = (
            ("list_vatrow", "Can list vat row"),
        )

    def __str__(self):
        return '{} {} {} {}'.format(self.order, self.vat, self.total_income, self.total_vat)

    def total(self):
        return self.total_income + self.total_fee - self.total_discount + self.total_vat

    def update_totals(self):
        order = self.order
        try:
            first_order_row = order.order_rows.first()
            first_order_row.update_totals()
            return True
        except Exception as e:
            return False, e
